@@ -33,14 +33,9 @@
 # DOCUMENTATION, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 # ===----------------------------------------------------------------------===
 
-# Python standard library, iteration tools
-from itertools import count
-
 # Haiku language, type definitions
 from haiku.types import *
 
-from haiku.utils.serialization import i2bytearray, i2varnumber, s2varstring
-
 from .base import BasePickler
 
 __all__ = [
@@ -61,6 +56,18 @@ __all__ = [
 #
 # That's it!
 
+# Python standard library,
+from copy import copy
+
+# Python standard library, iteration tools
+from itertools import count, izip
+
+# LEPL: Recursive descent parser for Python applications
+import lepl
+from lepl.core.dynamic import IntVar as lepl_IntVar
+
+from haiku.utils.serialization import i2bytearray, i2varnumber, s2varstring
+
 class CanonicalExpressionPickler(BasePickler):
   ""
   ASSOCIATION_OPERATOR = u"="
@@ -86,15 +93,20 @@ class CanonicalExpressionPickler(BasePickler):
     # convert it to s-expression notation.
     return self._serialize(args[0])
 
-  def load(self, istream):
-    return []
+  def loads(self, expression):
+    """Deserializes a haiku expression from a Unicode represented string in
+    “Canonical Expression” notation to Python objects."""
+    return self._matcher.parse(expression)
 
   def _serialize(self, expression):
     """Translates a Python-represented haiku expression into a byte string
     in “Canonical Expression” notation."""
     # None/nil/omega value:
     if isinstance(expression, OmegaCompatible):
-      return s2varstring('')
+      return ''.join([
+        self.TUPLE_OPEN.encode('utf-8'),
+        s2varstring('nil'),
+        self.TUPLE_CLOSE.encode('utf-8')])
 
     # Boolean literals:
     elif isinstance(expression, BooleanCompatible):
@@ -114,6 +126,7 @@ class CanonicalExpressionPickler(BasePickler):
       return ''.join([
         self.TUPLE_OPEN.encode('utf-8'),
         s2varstring('integer'),
+        self.QUOTE_OPERATOR.encode('utf-8'),
         s2varstring(expression and i2bytearray(expression) or ''),
         self.TUPLE_CLOSE.encode('utf-8')])
 
@@ -130,8 +143,14 @@ class CanonicalExpressionPickler(BasePickler):
     elif isinstance(expression, UnicodeCompatible):
       return ''.join([
         self.TUPLE_OPEN.encode('utf-8'),
-        s2varstring('string'),
+        s2varstring('decode'),
+        self.QUOTE_OPERATOR.encode('utf-8'),
         s2varstring(expression.encode('utf-8')),
+        self.ASSOCIATION_OPERATOR.encode('utf-8'),
+        self.QUOTE_OPERATOR.encode('utf-8'),
+        s2varstring('encoding'),
+        self.QUOTE_OPERATOR.encode('utf-8'),
+        s2varstring('utf-8'),
         self.TUPLE_CLOSE.encode('utf-8')])
 
     # Byte-array literals:
@@ -155,14 +174,15 @@ class CanonicalExpressionPickler(BasePickler):
       # Output special forms:
       if special_pattern(expression):
         if expression[0] == self.QUOTE_PROCEDURE:
-          #if (isinstance(expression[1], TupleCompatible) and
-          #    all(special_pattern(expression[1][key]) and
-          #        expression[1][key][0] == self.UNQUOTE_PROCEDURE
-          #        for key in expression[1].keys())):
-          #  return ''.join([
-          #    self.EVAL_DATA_OPEN.encode('utf-8'),
-          #    self.dumps(expression[1]),
-          #    self.EVAL_DATA_CLOSE.encode('utf-8')])
+          if (isinstance(expression[1], TupleCompatible) and
+              all(special_pattern(expression[1][key]) and
+                  expression[1][key][0] == self.UNQUOTE_PROCEDURE
+                  for key in expression[1])):
+            return ''.join([
+              self.EVAL_DATA_OPEN.encode('utf-8'),
+              self._serialize(Tuple((key, expression[1][key][1])
+                                    for key in expression[1]))[1:-1],
+              self.EVAL_DATA_CLOSE.encode('utf-8')])
           return ''.join([
             self.QUOTE_OPERATOR.encode('utf-8'),
             self._serialize(expression[1])])
@@ -221,6 +241,110 @@ class CanonicalExpressionPickler(BasePickler):
     raise ValueError(
       u"unrecognized input (not a valid expression): '%s'" % repr(expression))
 
+  def __init__(self, *args, **kwargs):
+    "Sets up a parser using the LEPL package."
+    super(CanonicalExpressionPickler, self).__init__(*args, **kwargs)
+
+    Expression = lepl.Delayed()
+
+    _UnsignedInteger = lambda digits:int(''.join(digits) or '0')
+    UnsignedInteger = (
+        lepl.Optional(lepl.Any('123456789') &
+                      lepl.Any('0123456789')[:]) &
+        ~lepl.Literal(':')
+      ) > _UnsignedInteger
+
+    _ByteArray = lambda bytes:Symbol(''.join(bytes))
+    _ByteArrayLength = lepl_IntVar()
+    _ByteArrayHeader = lepl.Apply(UnsignedInteger, _ByteArrayLength.setter())
+    _ByteArrayBody   = lepl.Repeat(lepl.Any(), stop=_ByteArrayLength, add_=True)
+    ByteArray = (
+      ~_ByteArrayHeader &
+      _ByteArrayBody) > _ByteArray
+    ByteArray.config.no_compile_to_regexp()
+
+    # Special forms for quoting:
+    _QuoteSyntax = lambda expr:Tuple([
+      (0, self.QUOTE_PROCEDURE),
+      (1, expr)])
+    QuoteSyntax = (
+      ~lepl.Literal(self.QUOTE_OPERATOR) & Expression) >> _QuoteSyntax
+
+    _UnquoteSyntax = lambda expr:Tuple([
+      (0, self.UNQUOTE_PROCEDURE),
+      (1, expr)])
+    UnquoteSyntax = (
+      ~lepl.Literal(self.UNQUOTE_OPERATOR) & Expression) >> _UnquoteSyntax
+
+    _UnquoteSpliceSyntax = lambda expr:Tuple([
+      (0, self.UNQUOTE_SPLICE_PROCEDURE),
+      (1, expr)])
+    UnquoteSpliceSyntax = (
+      ~lepl.Literal(self.UNQUOTE_SPLICE_OPERATOR) & Expression) >> _UnquoteSyntax
+
+    # A keyword expression is a component of the tuple definition: a mapping
+    # of one data to another (the key/value pair).
+    _KeywordExpression = lambda parts:len(parts)-1 and tuple(parts) or parts[0]
+    KeywordExpression = ((
+        ~lepl.Literal(self.ASSOCIATION_OPERATOR.encode('utf-8'))
+        & Expression & Expression
+      ) | Expression) > _KeywordExpression
+
+    # The creation of tuple values is a little tricky as keys may be
+    # specified either implicitly (by position) or explicitly.
+    def icount(*args, **kwargs):
+      _iter = count(*args, **kwargs)
+      for i in _iter:
+        yield Integer(i)
+    def _TupleSyntax(parts):
+      # (Remember, Python's `tuple` is quite different from haiku's `Tuple`)
+      kwargs = filter(lambda arg:isinstance(arg, tuple), parts)
+      args   = filter(lambda arg:arg not in kwargs, parts)
+      tuple_ = FrozenTuple(kwargs)
+      if len(kwargs) != len(tuple_):
+        raise self.SyntaxError(
+          u"duplicate keys in keyword arguments")
+      tuple_ = FrozenTuple([x for x in izip(icount(), args)] + kwargs)
+      if len(parts) != len(tuple_):
+        dups = filter(lambda x:x in tuple_, xrange(len(args)))
+        raise self.SyntaxError(
+          u"redundant parameter(s) specified positionally and as keyword arguments")
+      return tuple_
+    TupleSyntax = (
+      ~lepl.Literal(self.TUPLE_OPEN) &
+      KeywordExpression[0:] &
+      ~lepl.Literal(self.TUPLE_CLOSE)) > _TupleSyntax
+
+    # Eval-data special form:
+    def _EvalDataSyntax(parts):
+      tuple_ = _TupleSyntax(parts)
+      tuple_ = Tuple([(key, _UnquoteSyntax(value)) for key, value in tuple_.items()])
+      tuple_ = Tuple([(0, self.QUOTE_PROCEDURE), (1, tuple_)])
+      return tuple_
+    EvalDataSyntax = (
+      ~lepl.Literal(self.EVAL_DATA_OPEN) &
+      KeywordExpression[0:] &
+      ~lepl.Literal(self.EVAL_DATA_CLOSE)) > _EvalDataSyntax
+
+    # Sequence special form:
+    _SequenceSyntax = lambda args:Sequence(args)
+    SequenceSyntax = (
+      ~lepl.Literal(self.SEQUENCE_OPEN) &
+      Expression[0:] &
+      ~lepl.Literal(self.SEQUENCE_CLOSE)) > _SequenceSyntax
+
+    # Now that we've defined each component, we can go back and complete
+    # Expression's definition:
+    Expression += (QuoteSyntax | UnquoteSyntax | UnquoteSpliceSyntax |
+      TupleSyntax | EvalDataSyntax | SequenceSyntax | ByteArray)
+
+    # ...and our overall grammar: zero or more Expression's optionally
+    # separated by whitespace.
+    Syntax = Expression[0:] & ~lepl.Eos()
+
+    # Save the `Syntax` matcher for use by other methods:
+    self._matcher = Syntax
+
 # ===----------------------------------------------------------------------===
 # End of File
 # ===----------------------------------------------------------------------===
